@@ -92,8 +92,7 @@ semipublic_mod! {
     mod reload_cfg;
     mod socks;
 }
-
-#[cfg(feature = "rpc")]
+#[cfg_attr(not(feature = "rpc"), path = "rpc_stub.rs")]
 mod rpc;
 
 use std::ffi::OsString;
@@ -110,7 +109,7 @@ use arti_client::TorClient;
 use safelog::with_safe_logging_suppressed;
 use tor_config::mistrust::BuilderExt as _;
 use tor_config::ConfigurationSources;
-use tor_rtcompat::Runtime;
+use tor_rtcompat::ToplevelRuntime;
 
 use anyhow::{Context, Error, Result};
 use clap::{value_parser, Arg, ArgAction, Command};
@@ -125,14 +124,9 @@ use clap::Subcommand as _;
 pub use subcommands::proxy::run_proxy as run;
 
 /// Create a runtime for Arti to use.
-fn create_runtime() -> std::io::Result<impl Runtime> {
+fn create_runtime() -> std::io::Result<impl ToplevelRuntime> {
     cfg_if::cfg_if! {
-        if #[cfg(feature="rpc")] {
-            // TODO RPC: Because of
-            // https://gitlab.torproject.org/tpo/core/arti/-/issues/837 , we can
-            // currently define our RPC methods on TorClient<PreferredRuntime>.
-            use tor_rtcompat::PreferredRuntime as ChosenRuntime;
-        } else if #[cfg(all(feature="tokio", feature="native-tls"))] {
+        if #[cfg(all(feature="tokio", feature="native-tls"))] {
             use tor_rtcompat::tokio::TokioNativeTlsRuntime as ChosenRuntime;
         } else if #[cfg(all(feature="tokio", feature="rustls"))] {
             use tor_rtcompat::tokio::TokioRustlsRuntime as ChosenRuntime;
@@ -196,7 +190,7 @@ where
         // If we couldn't resolve the default config file, then too bad.  If something
         // actually tries to use it, it will produce an error, but don't fail here
         // just for that reason.
-        write!(config_file_help, " Defaults to {:?}", default).unwrap();
+        write!(config_file_help, " Defaults to {:?}", default).expect("Can't write to string");
     }
 
     // We create the runtime now so that we can use its `Debug` impl to describe it for
