@@ -26,8 +26,7 @@ use tor_socksproto::{Handshake as _, SocksAddr, SocksAuth, SocksCmd, SocksReques
 
 use anyhow::{anyhow, Context, Result};
 
-#[cfg(feature = "rpc")]
-use crate::rpc::RpcStateSender;
+use crate::rpc::RpcProxySupport;
 
 /// Payload to return when an HTTP connection arrive on a Socks port
 const WRONG_PROTOCOL_PAYLOAD: &[u8] = br#"HTTP/1.0 501 Tor is not an HTTP Proxy
@@ -148,7 +147,7 @@ struct AuthInterpretation {
 /// the [`ConnectWithPrefs`](arti_client::rpc::ConnectWithPrefs) special method.
 ///
 /// Right now, there are two general kinds of objects that implement this method:
-/// client-like objects, and stream-like objects.
+/// client-like objects, and one-shot clients.
 ///
 /// A client-like object is either a `TorClient` or an RPC `Session`.
 /// It knows about and it is capable of opening multiple data streams.
@@ -159,19 +158,14 @@ struct AuthInterpretation {
 /// An application gets a TorClient by asking the session for one,
 /// or for asking a TorClient to give you a new variant clone of itself.
 ///
-/// A stream-like object is an `arti_rpcserver::stream::RpcDataStream`.
-/// It is created from a client-like object, but represents a single data stream.
+/// A one-shot client is an `arti_rpcserver::stream::OneshotClient`.
+/// It is created from a client-like object, but can only be used for a single data stream.
 /// When created, it it not yet connected or trying to connect to anywhere:
 /// the act of using it as the target Object for a SOCKS connection causes
 /// it to begin connecting.
-/// (You can also think of this as a single-use client,
-/// which once used, becomes interchangeable with the DataStream it created.)
-/// (TODO: We may wish to change this vocabulary.
-/// We may wish to call this a "stream handle", for instance?)
 ///
-/// An application gets an RpcDataStream by calling `arti:new_stream_handle
-/// on any client-like object.  Currently, this always creates an RpcDataStream
-/// that makes optimistic connections; See #1583.
+/// An application gets a `OneShotClient` by calling `arti:new_oneshot_client`
+/// on any client-like object.
 ///
 /// ## The SOCKS protocol
 ///
@@ -250,7 +244,7 @@ struct AuthInterpretation {
 /// where the Session ID is `SESSION-1`.
 ///
 /// The application sends an RPC request of the form:
-/// `{"id": 123, "obj": "SESSION-1", "method": "arti:new_stream_handle", "params": {}}`
+/// `{"id": 123, "obj": "SESSION-1", "method": "arti:new_oneshot_client", "params": {}}`
 ///
 /// It receives a reply like:
 /// `{"id": 123, "result": {"id": "STREAM-1"} }`
@@ -438,7 +432,7 @@ impl<R: Runtime> SocksConnContext<R> {
 
         let client = self.tor_client.clone();
         #[cfg(feature = "rpc")]
-        let client = ConnTarget::Client(client);
+        let client = ConnTarget::Client(Box::new(client));
 
         Ok((prefs, client))
     }
@@ -450,6 +444,7 @@ impl<R: Runtime> SocksConnContext<R> {
 /// Uses `isolation_info` to decide which circuits this connection
 /// may use.  Requires that `isolation_info` is a pair listing the listener
 /// id and the source address for the socks request.
+#[allow(clippy::cognitive_complexity)] // TODO: Refactor
 async fn handle_socks_conn<R, S>(
     runtime: R,
     context: SocksConnContext<R>,
@@ -495,11 +490,12 @@ where
 
         match step {
             NS::Recv(mut recv) => {
-                let n = socks_stream.read(recv.buf())
+                let n = socks_stream
+                    .read(recv.buf())
                     .await
                     .context("Error while reading SOCKS handshake")?;
                 recv.note_received(n)?;
-            },
+            }
             NS::Send(data) => write_all_and_flush(&mut socks_stream, &data).await?,
             NS::Finished(fin) => break fin.into_output_forbid_pipelining()?,
         }
@@ -792,20 +788,19 @@ fn accept_err_is_fatal(err: &IoError) -> bool {
 /// timeouts, and a `tor_client` to use in connecting over the Tor
 /// network.
 #[cfg_attr(feature = "experimental-api", visibility::make(pub))]
+#[allow(clippy::cognitive_complexity)] // TODO: Refactor
 pub(crate) async fn run_socks_proxy<R: Runtime>(
     runtime: R,
     tor_client: TorClient<R>,
     listen: Listen,
-    // TODO RPC: This is not a good way to make an API conditional. We MUST
-    // refactor this before the RPC feature becomes non-experimental.
-    #[cfg(feature = "rpc")] rpc_data: Option<(
-        Arc<arti_rpcserver::RpcMgr>, //
-        RpcStateSender,
-    )>,
+    rpc_data: Option<RpcProxySupport>,
 ) -> Result<()> {
     #[cfg(feature = "rpc")]
     let (rpc_mgr, mut rpc_state_sender) = match rpc_data {
-        Some((m, s)) => (Some(m), Some(s)),
+        Some(RpcProxySupport {
+            rpc_mgr,
+            rpc_state_sender,
+        }) => (Some(rpc_mgr), Some(rpc_state_sender)),
         None => (None, None),
     };
 
