@@ -34,11 +34,13 @@
 #![deny(clippy::unnecessary_wraps)]
 #![warn(clippy::unseparated_literal_suffix)]
 #![deny(clippy::unwrap_used)]
+#![deny(clippy::mod_module_files)]
 #![allow(clippy::let_unit_value)] // This can reasonably be done for explicitness
 #![allow(clippy::uninlined_format_args)]
 #![allow(clippy::significant_drop_in_scrutinee)] // arti/-/merge_requests/588/#note_2812945
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
+#![allow(clippy::needless_lifetimes)] // See arti#1765
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 use std::path::Path;
@@ -118,10 +120,7 @@ impl LockFileGuard {
         if os::lockfile_has_path(&self.locked, path)? {
             std::fs::remove_file(path)
         } else {
-            Err(std::io::Error::new(
-                std::io::ErrorKind::Other,
-                MismatchedPathError {},
-            ))
+            Err(std::io::Error::other(MismatchedPathError {}))
         }
     }
 }
@@ -316,9 +315,9 @@ mod tests {
         test_temp_dir!().used_by(|dir| {
             let file = dir.join("file");
             let flock_guard = LockFileGuard::lock(&file).unwrap();
-            assert!(file.exists());
+            assert!(file.try_exists().unwrap());
             drop(flock_guard);
-            assert!(file.exists());
+            assert!(file.try_exists().unwrap());
         });
     }
 
@@ -327,9 +326,9 @@ mod tests {
         test_temp_dir!().used_by(|dir| {
             let file = dir.join("file");
             let flock_guard = LockFileGuard::lock(&file).unwrap();
-            assert!(file.exists());
+            assert!(file.try_exists().unwrap());
             assert!(flock_guard.delete_lock_file(&file).is_ok());
-            assert!(!file.exists());
+            assert!(!file.try_exists().unwrap());
         });
     }
 }
