@@ -126,7 +126,7 @@ impl StrExt for str {}
 pub trait RngExt: Rng {
     /// Generate a random value in the given range.
     ///
-    /// This function is optimised for the case that only a single sample is made from the given range. See also the [`Uniform`](rand::distributions::uniform::Uniform)  distribution type which may be faster if sampling from the same range repeatedly.
+    /// This function is optimised for the case that only a single sample is made from the given range. See also the [`Uniform`](rand::distr::uniform::Uniform)  distribution type which may be faster if sampling from the same range repeatedly.
     ///
     /// If the supplied range is empty, returns `None`.
     ///
@@ -135,7 +135,6 @@ pub trait RngExt: Rng {
     /// ### Example
     ///
     /// ```
-    /// use rand::thread_rng;
     /// use tor_basic_utils::RngExt as _;
     //
     // Fake plastic imitation tor_error, since that's actually higher up the stack
@@ -150,7 +149,7 @@ pub trait RngExt: Rng {
     /// use tor_error::{Bug, internal};
     ///
     /// fn choose(slice: &[i32]) -> Result<i32, Bug> {
-    ///     let index = thread_rng()
+    ///     let index = rand::rng()
     ///         .gen_range_checked(0..slice.len())
     ///         .ok_or_else(|| internal!("empty slice"))?;
     ///     Ok(slice[index])
@@ -159,16 +158,20 @@ pub trait RngExt: Rng {
     /// assert_eq!(choose(&[42]).unwrap(), 42);
     /// let _: Bug = choose(&[]).unwrap_err();
     /// ```
+    //
+    // TODO: We may someday wish to rename this function to random_range_checked,
+    // since gen_range was renamed to random_range in rand 0.9.
+    // Or we might decide to leave it alone.
     fn gen_range_checked<T, R>(&mut self, range: R) -> Option<T>
     where
-        T: rand::distributions::uniform::SampleUniform,
-        R: rand::distributions::uniform::SampleRange<T>,
+        T: rand::distr::uniform::SampleUniform,
+        R: rand::distr::uniform::SampleRange<T>,
     {
         if range.is_empty() {
             None
         } else {
             #[allow(clippy::disallowed_methods)]
-            Some(Rng::gen_range(self, range))
+            Some(Rng::random_range(self, range))
         }
     }
 
@@ -178,17 +181,16 @@ pub trait RngExt: Rng {
     /// with types that implement `GenRangeInfallible`
     /// (that necessarily then implement the appropriate `rand` traits).
     ///
-    /// This function is optimised for the case that only a single sample is made from the given range. See also the [`Uniform`](rand::distributions::uniform::Uniform)  distribution type which may be faster if sampling from the same range repeatedly.
+    /// This function is optimised for the case that only a single sample is made from the given range. See also the [`Uniform`](rand::distr::uniform::Uniform)  distribution type which may be faster if sampling from the same range repeatedly.
     ///
     /// ### Example
     ///
     /// ```
     /// use std::time::Duration;
-    /// use rand::thread_rng;
     /// use tor_basic_utils::RngExt as _;
     ///
     /// fn stochastic_sleep(max: Duration) {
-    ///     let chosen_delay = thread_rng()
+    ///     let chosen_delay = rand::rng()
     ///         .gen_range_infallible(..=max);
     ///     std::thread::sleep(chosen_delay);
     /// }
@@ -217,9 +219,9 @@ impl<T: Rng> RngExt for T {}
 // upper bound might be zero, unless a NonZero type is used, which seems like a further
 // complication that we probably don't want to introduce here.  That leaves lower-bounded
 // ranges, but those are very rare.
-pub trait GenRangeInfallible: rand::distributions::uniform::SampleUniform + Ord
+pub trait GenRangeInfallible: rand::distr::uniform::SampleUniform + Ord
 where
-    RangeInclusive<Self>: rand::distributions::uniform::SampleRange<Self>,
+    RangeInclusive<Self>: rand::distr::uniform::SampleRange<Self>,
 {
     /// The usual lower bound, for converting a `RangeToInclusive` to a `RangeInclusive`
     ///
@@ -589,6 +591,20 @@ macro_rules! derive_serde_raw { {
 
 // ----------------------------------------------------------------------
 
+/// Flatten a `Result<Result<T, E>, E>` into a `Result<T, E>`.
+///
+/// See the nightly [`Result::flatten`].
+// TODO MSRV TBD: When `Result::flatten` is stable and our MSRV allows,
+// remove this function and replace uses with `Result::flatten`.
+pub fn flatten<T, E>(x: Result<Result<T, E>, E>) -> Result<T, E> {
+    match x {
+        Ok(Ok(x)) => Ok(x),
+        Err(e) | Ok(Err(e)) => Err(e),
+    }
+}
+
+// ----------------------------------------------------------------------
+
 #[cfg(test)]
 mod test {
     // @@ begin test lint list maintained by maint/add_warning @@
