@@ -96,17 +96,21 @@ pub(crate) trait AbstractCirc: Debug {
 
     /// Return a [`Path`] object describing all the hops in this circuit.
     ///
+    /// Returns an error if the circuit is closed.
+    ///
     /// Note that this `Path` is not automatically updated if the circuit is
     /// extended.
-    fn path_ref(&self) -> Arc<Path>;
+    fn path_ref(&self) -> tor_proto::Result<Arc<Path>>;
 
     /// Return the number of hops in this circuit.
     ///
+    /// Returns an error if the circuit is closed.
+    ///
     /// NOTE: This function will currently return only the number of hops
     /// _currently_ in the circuit. If there is an extend operation in progress,
     /// the currently pending hop may or may not be counted, depending on whether
     /// the extend operation finishes before this call is done.
-    fn n_hops(&self) -> usize;
+    fn n_hops(&self) -> tor_proto::Result<usize>;
 
     /// Return true if this circuit is closed and therefore unusable.
     fn is_closing(&self) -> bool;
@@ -114,8 +118,8 @@ pub(crate) trait AbstractCirc: Debug {
     /// Return a process-unique identifier for this circuit.
     fn unique_id(&self) -> UniqId;
 
-    /// Extend the circuit via the ntor handshake to a new target last hop.
-    async fn extend_ntor<T: CircTarget + Sync>(
+    /// Extend the circuit via the most appropriate handshake to a new `target` hop.
+    async fn extend<T: CircTarget + Sync>(
         &self,
         target: &T,
         params: CircParameters,
@@ -784,6 +788,7 @@ impl<B: AbstractCircBuilder<R> + 'static, R: Runtime> AbstractCircMgr<B, R> {
     /// under the assumption that it will be used for that spec.
     ///
     /// This is the primary entry point for AbstractCircMgr.
+    #[allow(clippy::cognitive_complexity)] // TODO #2010: Refactor?
     pub(crate) async fn get_or_launch(
         self: &Arc<Self>,
         usage: &TargetCircUsage,
@@ -1003,6 +1008,7 @@ impl<B: AbstractCircBuilder<R> + 'static, R: Runtime> AbstractCircMgr<B, R> {
 
     /// Execute an action returned by pick-action, and return the
     /// resulting circuit or error.
+    #[allow(clippy::cognitive_complexity)] // TODO #2010: Refactor
     async fn take_action(
         self: Arc<Self>,
         act: Action<B, R>,
