@@ -34,11 +34,13 @@
 #![deny(clippy::unnecessary_wraps)]
 #![warn(clippy::unseparated_literal_suffix)]
 #![deny(clippy::unwrap_used)]
+#![deny(clippy::mod_module_files)]
 #![allow(clippy::let_unit_value)] // This can reasonably be done for explicitness
 #![allow(clippy::uninlined_format_args)]
 #![allow(clippy::significant_drop_in_scrutinee)] // arti/-/merge_requests/588/#note_2812945
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
+#![allow(clippy::needless_lifetimes)] // See arti#1765
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 use std::collections::HashMap;
@@ -57,6 +59,13 @@ use std::ffi::OsStr;
 #[cfg(feature = "address")]
 pub mod addr;
 
+#[cfg(feature = "arti-client")]
+mod arti_client_paths;
+
+#[cfg(feature = "arti-client")]
+#[cfg_attr(docsrs, doc(cfg(feature = "arti-client")))]
+pub use arti_client_paths::arti_client_base_resolver;
+
 /// A path in a configuration file: tilde expansion is performed, along
 /// with expansion of variables provided by a [`CfgPathResolver`].
 ///
