@@ -128,7 +128,15 @@ where
 
     // TODO(nickm) This should be an option, and is too long.
     let begin_timeout = Duration::from_secs(5);
-    let source = SourceInfo::from_circuit(&circuit);
+    let source = match SourceInfo::from_circuit(&circuit) {
+        Ok(source) => source,
+        Err(e) => {
+            return Err(Error::RequestFailed(RequestFailedError {
+                source: None,
+                error: e.into(),
+            }));
+        }
+    };
 
     let wrap_err = |error| {
         Error::RequestFailed(RequestFailedError {
@@ -137,7 +145,7 @@ where
         })
     };
 
-    req.check_circuit(&circuit).map_err(wrap_err)?;
+    req.check_circuit(&circuit).await.map_err(wrap_err)?;
 
     // Launch the stream.
     let mut stream = runtime
@@ -525,7 +533,10 @@ mod test {
     #![allow(clippy::needless_pass_by_value)]
     //! <!-- @@ end test lint list maintained by maint/add_warning @@ -->
     use super::*;
-    use tor_rtmock::{io::stream_pair, time::MockSleepProvider};
+    use tor_rtmock::io::stream_pair;
+
+    #[allow(deprecated)] // TODO #1885
+    use tor_rtmock::time::MockSleepProvider;
 
     use futures_await_test::async_test;
 
@@ -565,6 +576,7 @@ mod test {
     ) -> (RequestResult<()>, Vec<u8>) {
         // We don't need to do anything fancy here, since we aren't simulating
         // a timeout.
+        #[allow(deprecated)] // TODO #1885
         let mock_time = MockSleepProvider::new(std::time::SystemTime::now());
 
         let mut output = Vec::new();
@@ -771,7 +783,7 @@ mod test {
 
         let request = request?;
         assert!(request[..].starts_with(
-            b"GET /tor/micro/d/CQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQk.z HTTP/1.0\r\n"
+            b"GET /tor/micro/d/CQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQk HTTP/1.0\r\n"
         ));
 
         let response = response.unwrap();
