@@ -14,7 +14,7 @@ use tor_basic_utils::iter::{FilterCount, IteratorExt as _};
 use tor_linkspec::{ByRelayIds, HasRelayIds};
 
 use itertools::Itertools;
-use rand::seq::SliceRandom;
+use rand::seq::IndexedRandom;
 use serde::{Deserialize, Serialize};
 use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
@@ -334,7 +334,7 @@ impl GuardSet {
     }
 
     /// Return `Ok(true)` if `id` is definitely a member of this set, and
-    /// `Ok(false)` if it is definitely not a member.  
+    /// `Ok(false)` if it is definitely not a member.
     ///
     /// If we cannot tell, it's because there is a guard in this sample that has
     /// a _subset_ of the IDs in `id`. In that case, we return
@@ -922,7 +922,7 @@ impl GuardSet {
             options.truncate(1);
         }
 
-        match options.choose(&mut rand::thread_rng()) {
+        match options.choose(&mut rand::rng()) {
             Some((src, g)) => Ok((*src, g.guard_id().clone())),
             None => {
                 let retry_at = if running.n_accepted == 0 {
@@ -987,7 +987,7 @@ pub(crate) struct GuardSample<'a> {
     /// Equivalent to `GuardSet.guards.values()`, except in sample order.
     guards: Vec<Cow<'a, Guard>>,
     /// The identities for the confirmed members of `guards`, in confirmed order.
-    confirmed: Cow<'a, Vec<GuardId>>,
+    confirmed: Cow<'a, [GuardId]>,
     /// Other data from the state file that this version of Arti doesn't recognize.
     #[serde(flatten)]
     remaining: HashMap<String, JsonValue>,
@@ -1048,7 +1048,7 @@ mod test {
         // has 20 viable (Guard + V2Dir + DirCache=2) candidates.  Otherwise the
         // calculation of collision probability at the end of this function is
         // too tricky.
-        let netdir = tor_netdir::testnet::construct_custom_netdir(|idx, builder| {
+        let netdir = tor_netdir::testnet::construct_custom_netdir(|idx, builder, _| {
             // Give every node equal bandwidth.
             builder.rs.weight(RelayWeight::Measured(1000));
             // The default network has 40 relays, and the first 10 are
@@ -1478,7 +1478,7 @@ mod test {
         assert_eq!(guards.n_primary_without_id_info_in(&netdir), 0);
 
         use tor_netdir::testnet;
-        let netdir2 = testnet::construct_custom_netdir(|_idx, bld| {
+        let netdir2 = testnet::construct_custom_netdir(|_idx, bld, _| {
             let md_so_far = bld.md.testing_md().expect("Couldn't build md?");
             if &p_id1.0.identity(RelayIdType::Ed25519).unwrap() == md_so_far.ed25519_id() {
                 bld.omit_md = true;
