@@ -431,6 +431,7 @@ impl<'c, R: Runtime, M: MocksForConnect<R>> Context<'c, R, M> {
     ///
     /// Does all necessary retries and timeouts.
     /// Returns an error if no valid descriptor could be found.
+    #[allow(clippy::cognitive_complexity)] // TODO: Refactor
     async fn descriptor_ensure<'d>(&self, data: &'d mut DataHsDesc) -> Result<&'d HsDesc, CE> {
         // Maximum number of hsdir connection and retrieval attempts we'll make
         let max_total_attempts = self
@@ -510,9 +511,10 @@ impl<'c, R: Runtime, M: MocksForConnect<R>> Context<'c, R, M> {
                 Err(error) => {
                     debug_report!(
                         &error,
-                        "failed hsdir desc fetch for {} from {}",
+                        "failed hsdir desc fetch for {} from {}/{}",
                         &self.hsid,
                         &relay.id(),
+                        &relay.rsa_id()
                     );
                     errors.push(tor_error::Report(DescriptorError {
                         hsdir: hsdir_for_error,
@@ -704,7 +706,7 @@ impl<'c, R: Runtime, M: MocksForConnect<R>> Context<'c, R, M> {
                     intro_index,
                     intro_desc,
                     intro_target,
-                    sort_rand: self.mocks.thread_rng().gen(),
+                    sort_rand: self.mocks.thread_rng().random(),
                 })
             })
             .filter_map(|entry| match entry {
@@ -978,7 +980,7 @@ impl<'c, R: Runtime, M: MocksForConnect<R>> Context<'c, R, M> {
         let rend_pt = rend_pt_identity_for_error(&rend_relay);
         *using_rend_pt = Some(rend_pt.clone());
 
-        let rend_cookie: RendCookie = self.mocks.thread_rng().gen();
+        let rend_cookie: RendCookie = self.mocks.thread_rng().random();
         let message = EstablishRendezvous::new(rend_cookie);
 
         let (rend_established_tx, rend_established_rx) = proto_oneshot::channel();
@@ -1059,6 +1061,7 @@ impl<'c, R: Runtime, M: MocksForConnect<R>> Context<'c, R, M> {
     /// So if there's a failure, it's purely to do with the introduction point.
     ///
     /// Does not apply a timeout.
+    #[allow(clippy::cognitive_complexity)] // TODO: Refactor
     async fn exchange_introduce(
         &'c self,
         ipt: &UsableIntroPt<'_>,
@@ -1421,7 +1424,7 @@ impl<R: Runtime> MocksForConnect<R> for () {
     type Rng = rand::rngs::ThreadRng;
 
     fn thread_rng(&self) -> Self::Rng {
-        rand::thread_rng()
+        rand::rng()
     }
 }
 #[async_trait]
@@ -1524,6 +1527,7 @@ mod test {
     use tor_netdoc::doc::{hsdesc::test_data, netstatus::Lifetime};
     use tor_rtcompat::tokio::TokioNativeTlsRuntime;
     use tor_rtcompat::RuntimeSubstExt as _;
+    #[allow(deprecated)] // TODO #1885
     use tor_rtmock::time::MockSleepProvider;
     use tracing_test::traced_test;
 
@@ -1647,6 +1651,7 @@ mod test {
         let netdir = Arc::new(netdir.unwrap_if_sufficient().unwrap());
         let runtime = TokioNativeTlsRuntime::current().unwrap();
         let now = humantime::parse_rfc3339("2023-02-09T12:00:00Z").unwrap();
+        #[allow(deprecated)] // TODO #1885
         let mock_sp = MockSleepProvider::new(now);
         let runtime = runtime
             .with_sleep_provider(mock_sp.clone())
