@@ -70,6 +70,10 @@ pub(crate) enum IptEstablisherError {
     #[error("Unable to build circuit to introduction point")]
     BuildCircuit(#[source] tor_circmgr::Error),
 
+    /// We encountered an error while querying the circuit state.
+    #[error("Unable to query circuit state")]
+    CircuitState(#[source] tor_proto::Error),
+
     /// We encountered an error while building and signing our establish_intro
     /// message.
     #[error("Unable to construct signed ESTABLISH_INTRO message")]
@@ -135,6 +139,7 @@ impl tor_error::HasKind for IptEstablisherError {
             E::Ipt(e) => e.kind(),
             E::NetdirProviderShutdown(e) => e.kind(),
             E::BuildCircuit(e) => e.kind(),
+            E::CircuitState(e) => e.kind(),
             E::EstablishTimeout => EK::TorNetworkTimeout,
             E::SendEstablishIntro(e) => e.kind(),
             E::ClosedWithoutAck => EK::CircuitCollapse,
@@ -176,6 +181,7 @@ impl IptEstablisherError {
             // This _might_ be the introduction point's fault, but it might not.
             // We can't be certain.
             IE::BuildCircuit(_) => None,
+            IE::CircuitState(_) => None,
             IE::EstablishTimeout => None,
             IE::ClosedWithoutAck => None,
             // These are, most likely, not the introduction point's fault,
@@ -230,7 +236,7 @@ pub(crate) struct IptParameters {
     ///
     /// We use this to record the requests that we see, and to prevent replays.
     #[educe(Debug(ignore))]
-    pub(crate) replay_log: ReplayLog,
+    pub(crate) replay_log: IptReplayLog,
     /// A set of identifiers for the Relay that we intend to use as the
     /// introduction point.
     ///
@@ -616,7 +622,7 @@ struct Reactor<R: Runtime> {
     ///
     /// Has to be an async mutex since it's locked for a long time,
     /// so we mustn't block the async executor thread on it.
-    replay_log: Arc<futures::lock::Mutex<ReplayLog>>,
+    replay_log: Arc<futures::lock::Mutex<IptReplayLog>>,
 }
 
 /// An open session with a single introduction point.
@@ -631,6 +637,7 @@ pub(crate) struct IntroPtSession {
 impl<R: Runtime> Reactor<R> {
     /// Run forever, keeping an introduction point established.
     #[allow(clippy::blocks_in_conditions)]
+    #[allow(clippy::cognitive_complexity)]
     async fn keep_intro_established(
         &self,
         mut status_tx: DropNotifyWatchSender<IptStatus>,
@@ -690,7 +697,7 @@ impl<R: Runtime> Reactor<R> {
                         &self.nickname,
                         self.target.display_relay_ids().redacted()
                     );
-                    let retry_after = retry_delay.next_delay(&mut rand::thread_rng());
+                    let retry_after = retry_delay.next_delay(&mut rand::rng());
                     self.runtime.sleep(retry_after).await;
                 }
             }
@@ -734,13 +741,15 @@ impl<R: Runtime> Reactor<R> {
             let mut details = EstablishIntroDetails::new(ipt_sid_id);
             if let Some(dos_params) = &self.extensions.dos_params {
                 // We only send the Dos extension when the relay is known to
-                // support HsIntro=5.
-                if protovers.supports_known_subver(tor_protover::ProtoKind::HSIntro, 5) {
+                // support it.
+                use tor_protover::named::HSINTRO_RATELIM;
+                if protovers.supports_named_subver(HSINTRO_RATELIM) {
                     details.set_extension_dos(dos_params.clone());
                 }
             }
             let circuit_binding_key = circuit
                 .binding_key(intro_pt_hop)
+                .map_err(IptEstablisherError::CircuitState)?
                 .ok_or(internal!("No binding key for introduction point!?"))?;
             let body: Vec<u8> = details
                 .sign_and_encode((*self.k_sid).as_ref(), circuit_binding_key.hs_mac())
@@ -765,7 +774,7 @@ impl<R: Runtime> Reactor<R> {
         let (established_tx, established_rx) = oneshot::channel();
 
         // In theory there ought to be only one IptMsgHandler in existence at any one time,
-        // for any one IptLocalId (ie for any one ReplayLog).  However, the teardown
+        // for any one IptLocalId (ie for any one IptReplayLog).  However, the teardown
         // arrangements are (i) complicated (so might have bugs) and (ii) asynchronous
         // (so we need to synchronise).  Therefore:
         //
@@ -790,11 +799,12 @@ impl<R: Runtime> Reactor<R> {
         // that the message was sent.  We have to wait for any actual `established`
         // message, though.
 
+        let length = circuit
+            .n_hops()
+            .map_err(into_internal!("failed to get circuit length"))?;
         let ack_timeout = self
             .pool
-            .estimate_timeout(&tor_circmgr::timeouts::Action::RoundTrip {
-                length: circuit.n_hops(),
-            });
+            .estimate_timeout(&tor_circmgr::timeouts::Action::RoundTrip { length });
         let _established: IntroEstablished = self
             .runtime
             .timeout(ack_timeout, established_rx)
@@ -844,7 +854,7 @@ struct IptMsgHandler {
     lid: IptLocalId,
 
     /// A replay log used to detect replayed introduction requests.
-    replay_log: futures::lock::OwnedMutexGuard<ReplayLog>,
+    replay_log: futures::lock::OwnedMutexGuard<IptReplayLog>,
 }
 
 impl tor_proto::circuit::MsgHandler for IptMsgHandler {
