@@ -1,5 +1,6 @@
 //! An error type for the `tor-keymgr` crate.
 
+use tor_basic_utils::PathExt;
 use tor_error::HasKind;
 
 use dyn_clone::DynClone;
@@ -7,6 +8,7 @@ use tor_persist::slug::BadSlug;
 
 use std::error::Error as StdError;
 use std::fmt;
+use std::path::PathBuf;
 use std::sync::Arc;
 
 use crate::KeyPathError;
@@ -38,6 +40,10 @@ pub enum Error {
     #[error("{0}")]
     KeyForge(#[from] tor_key_forge::Error),
 
+    /// An error caused by an invalid certificate.
+    #[error("{0}")]
+    InvalidCert(#[from] tor_key_forge::InvalidCertError),
+
     /// An internal error.
     #[error("Internal error")]
     Bug(#[from] tor_error::Bug),
@@ -59,6 +65,7 @@ impl HasKind for Error {
             E::Corruption(_) => EK::KeystoreCorrupted,
             E::KeyAlreadyExists => EK::BadApiUsage, // TODO: not strictly right
             E::KeyForge(_) => EK::BadApiUsage,
+            E::InvalidCert(_) => EK::BadApiUsage, // TODO: not strictly right
             E::Bug(e) => e.kind(),
         }
     }
@@ -71,13 +78,17 @@ impl HasKind for Error {
 ///
 /// (Does not include any errors arising from paths which are invalid
 /// *for the particular key*.)
-#[derive(thiserror::Error, Debug, Clone, Eq, PartialEq)]
+#[derive(thiserror::Error, Debug, Clone)]
 #[error("Invalid ArtiPath")]
 #[non_exhaustive]
 pub enum ArtiPathSyntaxError {
     /// One of the path slugs was invalid.
     #[error("{0}")]
     Slug(#[from] BadSlug),
+
+    /// An internal error.
+    #[error("Internal error")]
+    Bug(#[from] tor_error::Bug),
 }
 
 /// An error caused by keystore corruption.
@@ -88,6 +99,18 @@ pub enum KeystoreCorruptionError {
     /// A keystore contains a key that has an invalid [`KeyPath`](crate::KeyPath).
     #[error("{0}")]
     KeyPath(#[from] KeyPathError),
+
+    /// Missing certificate for key.
+    #[error("Missing certificate for key")]
+    MissingCertificate,
+
+    /// Missing the subject key of a certificate we own.
+    #[error("Subject key of certificate not found")]
+    MissingSubjectKey,
+
+    /// Missing signing key for certificate.
+    #[error("Missing signing key for certificate")]
+    MissingSigningKey,
 }
 
 /// An error that happens when we encounter an unknown key type.
@@ -98,6 +121,52 @@ pub struct UnknownKeyTypeError {
     pub(crate) arti_extension: String,
 }
 
+/// An unrecognized keystore entry.
+#[derive(Clone, Debug, amplify::Getters, thiserror::Error)]
+#[error("Unrecognized keystore entry {}", entry)]
+pub struct UnrecognizedEntryError {
+    /// An identifier of the entry that caused the error.
+    entry: UnrecognizedEntryId,
+    /// The underlying error that occurred.
+    // TODO: This should be an `Error` specific for the situation.
+    //
+    // [`KeystoreError`] is a provvisory solution that presents
+    // some issues, for example:
+    //
+    // * not all variants of `KeystoreError` are relevant
+    // * redundancy with some other Error types like
+    // [`MalformedServiceKeyError::NotAKey`](crate::keystore::ctor::err::MalformedServiceKeyError)
+    // * [`Keystore::list`](crate::Keystore) returns
+    // `StdResult<Vec<StdResult<(KeyPath, KeystoreItemType), UnrecognizedEntryError>>, KeystoreError>`,
+    // `KeystoreError` presents itself twice at 2 different levels, there is abiguity
+    #[source]
+    error: Arc<dyn KeystoreError>,
+}
+
+impl UnrecognizedEntryError {
+    /// Create a new instance of `KeystoreListError` given an `UnrecognizedEntryId`
+    /// and an `Arc<dyn KeystoreError>`.
+    pub fn new(entry: UnrecognizedEntryId, error: Arc<dyn KeystoreError>) -> Self {
+        Self { entry, error }
+    }
+}
+
+/// The identifier of an unrecognized key inside a [`Keystore`](crate::Keystore).
+///
+/// The exact type of the identifier depends on the backing storage of the keystore
+/// (for example, an on-disk keystore will identify its entries by [`Path`](UnrecognizedEntryId::Path)).
+#[non_exhaustive]
+#[derive(Debug, Clone, PartialEq, derive_more::Display)]
+pub enum UnrecognizedEntryId {
+    /// An entry identified by path inside an on-disk keystore.
+    // NOTE: this will only be used by on-disk keystores like
+    // [`ArtiNativeKeystore`](crate::ArtiNativeKeystore)
+    #[display("{}", _0.display_lossy())]
+    Path(PathBuf),
+    // TODO: when/if we add support for non on-disk keystores,
+    // new variants will be added
+}
+
 #[cfg(test)]
 mod tests {
     // @@ begin test lint list maintained by maint/add_warning @@
