@@ -194,7 +194,7 @@ impl HsDescInner {
     /// inner document, and the inner document itself.
     #[cfg_attr(feature = "hsdesc-inner-docs", visibility::make(pub))]
     pub(super) fn parse(s: &str) -> Result<(Option<Ed25519Identity>, UncheckedHsDescInner)> {
-        let mut reader = NetDocReader::new(s);
+        let mut reader = NetDocReader::new(s)?;
         let result = Self::take_from_reader(&mut reader).map_err(|e| e.within(s))?;
         Ok(result)
     }
@@ -496,7 +496,8 @@ mod test {
                 .1,
         );
         for n in NUM_INTRO_POINT_MAX..NUM_INTRO_POINT_MAX + 2 {
-            let many = chain!(iter::once(&*none), iter::repeat(&*ipt).take(n),).collect::<String>();
+            let many =
+                chain!(iter::once(&*none), std::iter::repeat_n(&*ipt, n),).collect::<String>();
             let desc = HsDescInner::parse(&many).unwrap();
             let desc = desc
                 .1
