@@ -9,7 +9,7 @@
 use super::circmap::{CircEnt, CircMap};
 use super::OpenChanCellS2C;
 use crate::channel::OpenChanMsgS2C;
-use crate::circuit::halfcirc::HalfCirc;
+use crate::tunnel::circuit::halfcirc::HalfCirc;
 use crate::util::err::ReactorError;
 use crate::util::oneshot_broadcast;
 use crate::{Error, Result};
@@ -18,7 +18,12 @@ use tor_cell::chancell::msg::{Destroy, DestroyReason, PaddingNegotiate};
 use tor_cell::chancell::ChanMsg;
 use tor_cell::chancell::{msg::AnyChanMsg, AnyChanCell, CircId};
 use tor_memquota::mq_queue;
-use tor_rtcompat::{SleepProvider, StreamOps};
+use tor_rtcompat::SleepProvider;
+
+#[cfg_attr(not(target_os = "linux"), allow(unused))]
+use tor_error::error_report;
+#[cfg_attr(not(target_os = "linux"), allow(unused))]
+use tor_rtcompat::StreamOps;
 
 use futures::channel::mpsc;
 use oneshot_fused_workaround as oneshot;
@@ -28,7 +33,7 @@ use futures::stream::Stream;
 use futures::Sink;
 use futures::StreamExt as _;
 use futures::{select, select_biased};
-use tor_error::{error_report, internal};
+use tor_error::internal;
 
 use std::fmt;
 use std::pin::Pin;
@@ -37,7 +42,7 @@ use std::sync::Arc;
 use crate::channel::{
     codec::CodecError, kist::KistParams, padding, params::*, unique_id, ChannelDetails, CloseInfo,
 };
-use crate::circuit::{celltypes::CreateResponse, CircuitRxSender};
+use crate::tunnel::circuit::{celltypes::CreateResponse, CircuitRxSender};
 use tracing::{debug, trace};
 
 /// A boxed trait object that can provide `ChanCell`s.
@@ -80,7 +85,7 @@ pub enum CtrlMsg {
         /// Channel to send other messages from this circuit down.
         sender: CircuitRxSender,
         /// Oneshot channel to send the new circuit's identifiers down.
-        tx: ReactorResultChannel<(CircId, crate::circuit::UniqId)>,
+        tx: ReactorResultChannel<(CircId, crate::tunnel::circuit::UniqId)>,
     },
     /// Enable/disable/reconfigure channel padding
     ///
@@ -123,6 +128,7 @@ pub struct Reactor<S: SleepProvider> {
     /// This should also be backed by a TLS connection if you want it to be secure.
     pub(super) output: BoxedChannelSink,
     /// A handler for setting stream options on the underlying stream.
+    #[cfg_attr(not(target_os = "linux"), allow(unused))]
     pub(super) streamops: BoxedChannelStreamOps,
     /// Timer tracking when to generate channel padding
     pub(super) padding_timer: Pin<Box<padding::Timer<S>>>,
@@ -275,7 +281,7 @@ impl<S: SleepProvider> Reactor<S> {
                 sender,
                 tx,
             } => {
-                let mut rng = rand::thread_rng();
+                let mut rng = rand::rng();
                 let my_unique_id = self.unique_id;
                 let circ_unique_id = self.circ_unique_id_ctx.next(my_unique_id);
                 let ret: Result<_> = self
@@ -301,7 +307,7 @@ impl<S: SleepProvider> Reactor<S> {
                     padding_negotiate,
                 } = &*updates;
                 if let Some(parameters) = padding_parameters {
-                    self.padding_timer.as_mut().reconfigure(parameters);
+                    self.padding_timer.as_mut().reconfigure(parameters)?;
                 }
                 if let Some(enable) = padding_enable {
                     if *enable {
@@ -509,8 +515,8 @@ pub(crate) mod test {
     #![allow(clippy::unwrap_used)]
     use super::*;
     use crate::channel::{ClosedUnexpectedly, UniqId};
-    use crate::circuit::CircParameters;
     use crate::fake_mpsc;
+    use crate::tunnel::circuit::CircParameters;
     use crate::util::fake_mq;
     use futures::sink::SinkExt;
     use futures::stream::StreamExt;
@@ -669,8 +675,7 @@ pub(crate) mod test {
                 reactor.run_once().await.unwrap();
             };
 
-            let (circ, _) =
-                futures::join!(pending.create_firsthop_fast(&circparams), send_response);
+            let (circ, _) = futures::join!(pending.create_firsthop_fast(circparams), send_response);
             // Make sure statuses are as expected.
             assert!(matches!(circ.err().unwrap(), Error::BadCircHandshakeAuth));
 
@@ -726,7 +731,7 @@ pub(crate) mod test {
     #[test]
     fn deliver_relay() {
         tor_rtcompat::test_with_all_runtimes!(|rt| async move {
-            use crate::circuit::celltypes::ClientCircChanMsg;
+            use crate::tunnel::circuit::celltypes::ClientCircChanMsg;
             use oneshot_fused_workaround as oneshot;
 
             let (_chan, mut reactor, _output, mut input) = new_reactor(rt);
@@ -814,7 +819,7 @@ pub(crate) mod test {
     #[test]
     fn deliver_destroy() {
         tor_rtcompat::test_with_all_runtimes!(|rt| async move {
-            use crate::circuit::celltypes::*;
+            use crate::tunnel::circuit::celltypes::*;
             use oneshot_fused_workaround as oneshot;
 
             let (_chan, mut reactor, _output, mut input) = new_reactor(rt);
