@@ -2,9 +2,9 @@
 //! for byte-oriented communication.
 
 use crate::{Error, Result};
-use futures::future::BoxFuture;
+use static_assertions::assert_impl_all;
 use tor_cell::relaycell::msg::EndReason;
-use tor_cell::relaycell::RelayCmd;
+use tor_cell::relaycell::{RelayCellFormat, RelayCmd};
 
 use futures::io::{AsyncRead, AsyncWrite};
 use futures::task::{Context, Poll};
@@ -29,11 +29,11 @@ use std::sync::{Mutex, Weak};
 use educe::Educe;
 
 #[cfg(any(feature = "experimental-api", feature = "stream-ctrl"))]
-use crate::circuit::ClientCirc;
+use crate::tunnel::circuit::ClientCirc;
 
-use crate::circuit::StreamTarget;
 use crate::memquota::StreamAccount;
 use crate::stream::StreamReader;
+use crate::tunnel::StreamTarget;
 use tor_basic_utils::skip_fmt;
 use tor_cell::relaycell::msg::Data;
 use tor_error::internal;
@@ -122,21 +122,12 @@ pub struct DataStream {
     w: DataWriter,
     /// Underlying reader for this stream
     r: DataReader,
-    /// A handle for the underlying circuit.
-    ///
-    /// TODO: This is redundant with the reference in `StreamTarget` inside
-    /// DataWriterState, but for now we can't actually access that state all the time,
-    /// since it might be inside a boxed future.
-    ///
-    /// TODO: This is also redundant with the reference in `ClientDataStreamCtrl`.
-    #[cfg(feature = "experimental-api")]
-    circuit: Arc<ClientCirc>,
-
     /// A control object that can be used to monitor and control this stream
     /// without needing to own it.
     #[cfg(feature = "stream-ctrl")]
     ctrl: std::sync::Arc<ClientDataStreamCtrl>,
 }
+assert_impl_all! { DataStream: Send, Sync }
 
 /// An object used to control and monitor a data stream.
 ///
@@ -157,6 +148,7 @@ pub struct ClientDataStreamCtrl {
     ///
     /// We make this a Weak reference so that once the stream itself is closed,
     /// we can't leak circuits.
+    // TODO(conflux): use ClientTunnel
     circuit: Weak<ClientCirc>,
 
     /// Shared user-visible information about the state of this stream.
@@ -332,6 +324,7 @@ restricted_msg! {
 // ClientDataStreamCtrl?
 #[cfg(feature = "stream-ctrl")]
 impl super::ctrl::ClientStreamCtrl for ClientDataStreamCtrl {
+    // TODO(conflux): use ClientTunnel
     fn circuit(&self) -> Option<Arc<ClientCirc>> {
         self.circuit.upgrade()
     }
@@ -381,6 +374,9 @@ impl DataStream {
         connected: bool,
         memquota: StreamAccount,
     ) -> Self {
+        let relay_cell_format = target.relay_cell_format();
+        let out_buf_len = Data::max_body_len(relay_cell_format);
+
         #[cfg(feature = "stream-ctrl")]
         let status = {
             let mut data_stream_status = DataStreamStatus::default();
@@ -389,14 +385,13 @@ impl DataStream {
             }
             Arc::new(Mutex::new(data_stream_status))
         };
+
         #[cfg(feature = "stream-ctrl")]
         let ctrl = Arc::new(ClientDataStreamCtrl {
             circuit: Arc::downgrade(target.circuit()),
             status: status.clone(),
             _memquota: memquota.clone(),
         });
-        #[cfg(feature = "experimental-api")]
-        let circuit = target.circuit().clone();
         let r = DataReader {
             state: Some(DataReaderState::Ready(DataReaderImpl {
                 s: reader,
@@ -413,10 +408,11 @@ impl DataStream {
         let w = DataWriter {
             state: Some(DataWriterState::Ready(DataWriterImpl {
                 s: target,
-                buf: Box::new([0; Data::MAXLEN]),
+                buf: vec![0; out_buf_len].into_boxed_slice(),
                 n_pending: 0,
                 #[cfg(feature = "stream-ctrl")]
                 status,
+                relay_cell_format,
             })),
             _memquota: memquota,
             #[cfg(feature = "stream-ctrl")]
@@ -425,8 +421,6 @@ impl DataStream {
         DataStream {
             w,
             r,
-            #[cfg(feature = "experimental-api")]
-            circuit,
             #[cfg(feature = "stream-ctrl")]
             ctrl,
         }
@@ -465,24 +459,6 @@ impl DataStream {
         }
     }
 
-    /// Return a reference to this stream's circuit?
-    ///
-    /// This is an experimental API; it is not covered by semver guarantee. It
-    /// is likely to change or disappear in a future release.
-    ///
-    /// TODO: Should there be an AttachedToCircuit trait that we use for all
-    /// client stream types?  Should this return an Option<&ClientCirc>?
-    ///
-    /// TODO RPC: Perhaps we should deprecate this in favor of `stream.ctrl().circuit()`.
-    #[cfg(feature = "experimental-api")]
-    #[deprecated(
-        since = "0.27.0",
-        note = "Use client_stream_ctrl()?.circuit()? instead."
-    )]
-    pub fn circuit(&self) -> &ClientCirc {
-        &self.circuit
-    }
-
     /// Return a [`ClientDataStreamCtrl`] object that can be used to monitor and
     /// interact with this stream without holding the stream itself.
     #[cfg(feature = "stream-ctrl")]
@@ -543,6 +519,9 @@ impl TokioAsyncWrite for DataStream {
     }
 }
 
+/// Helper type: Like BoxFuture, but also requires that the future be Sync.
+type BoxSyncFuture<'a, T> = Pin<Box<dyn Future<Output = T> + Send + Sync + 'a>>;
+
 /// An enumeration for the state of a DataWriter.
 ///
 /// We have to use an enum here because, for as long as we're waiting
@@ -558,8 +537,8 @@ enum DataWriterState {
     Ready(DataWriterImpl),
     /// The writer is flushing a cell.
     Flushing(
-        #[educe(Debug(method = "skip_fmt"))]
-        Pin<Box<dyn Future<Output = (DataWriterImpl, Result<()>)> + Send>>,
+        #[educe(Debug(method = "skip_fmt"))] //
+        BoxSyncFuture<'static, (DataWriterImpl, Result<()>)>,
     ),
 }
 
@@ -571,16 +550,23 @@ struct DataWriterImpl {
     s: StreamTarget,
 
     /// Buffered data to send over the connection.
+    ///
+    /// This buffer is currently allocated using a number of bytes
+    /// equal to the maximum that we can package at a time.
+    //
     // TODO: this buffer is probably smaller than we want, but it's good
     // enough for now.  If we _do_ make it bigger, we'll have to change
     // our use of Data::split_from to handle the case where we can't fit
     // all the data.
     #[educe(Debug(method = "skip_fmt"))]
-    buf: Box<[u8; Data::MAXLEN]>,
+    buf: Box<[u8]>,
 
     /// Number of unflushed bytes in buf.
     n_pending: usize,
 
+    /// Relay cell format in use
+    relay_cell_format: RelayCellFormat,
+
     /// Shared user-visible information about the state of this stream.
     #[cfg(feature = "stream-ctrl")]
     status: Arc<Mutex<DataStreamStatus>>,
@@ -604,7 +590,7 @@ impl DataWriter {
         let state = self.state.take().expect("Missing state in DataWriter");
 
         // TODO: this whole function is a bit copy-pasted.
-        let mut future: BoxFuture<_> = match state {
+        let mut future: BoxSyncFuture<_> = match state {
             DataWriterState::Ready(imp) => {
                 if imp.n_pending == 0 {
                     // Nothing to flush!
@@ -743,16 +729,17 @@ impl TokioAsyncWrite for DataWriter {
 impl DataWriterImpl {
     /// Try to flush the current buffer contents as a data cell.
     async fn flush_buf(mut self) -> (Self, Result<()>) {
-        let result =
-            if let Some((cell, remainder)) = Data::try_split_from(&self.buf[..self.n_pending]) {
-                // TODO: Eventually we may want a larger buffer; if we do,
-                // this invariant will become false.
-                assert!(remainder.is_empty());
-                self.n_pending = 0;
-                self.s.send(cell.into()).await
-            } else {
-                Ok(())
-            };
+        let result = if let Some((cell, remainder)) =
+            Data::try_split_from(self.relay_cell_format, &self.buf[..self.n_pending])
+        {
+            // TODO: Eventually we may want a larger buffer; if we do,
+            // this invariant will become false.
+            assert!(remainder.is_empty());
+            self.n_pending = 0;
+            self.s.send(cell.into()).await
+        } else {
+            Ok(())
+        };
 
         (self, result)
     }
@@ -800,8 +787,8 @@ enum DataReaderState {
     /// The reader is currently fetching a cell: this future is the
     /// progress it is making.
     ReadingCell(
-        #[educe(Debug(method = "skip_fmt"))]
-        Pin<Box<dyn Future<Output = (DataReaderImpl, Result<()>)> + Send>>,
+        #[educe(Debug(method = "skip_fmt"))] //
+        BoxSyncFuture<'static, (DataReaderImpl, Result<()>)>,
     ),
 }
 
