@@ -4,12 +4,12 @@ pub(super) mod create;
 pub(super) mod extender;
 
 use crate::channel::{Channel, ChannelSender};
-use crate::congestion::sendme::{self, CircTag};
+use crate::congestion::sendme;
 use crate::congestion::{CongestionControl, CongestionSignals};
 use crate::crypto::binding::CircuitBinding;
 use crate::crypto::cell::{
     HopNum, InboundClientCrypt, InboundClientLayer, OutboundClientCrypt, OutboundClientLayer,
-    RelayCellBody, SENDME_TAG_LEN,
+    RelayCellBody,
 };
 use crate::crypto::handshake::fast::CreateFastClient;
 use crate::crypto::handshake::ntor::{NtorClient, NtorPublicKey};
@@ -38,8 +38,8 @@ use tor_async_utils::{SinkTrySend as _, SinkTrySendError as _};
 use tor_cell::chancell::msg::{AnyChanMsg, HandshakeType, Relay};
 use tor_cell::chancell::{AnyChanCell, ChanCmd, CircId};
 use tor_cell::chancell::{BoxedCellBody, ChanMsg};
-use tor_cell::relaycell::extend::NtorV3Extension;
-use tor_cell::relaycell::msg::{AnyRelayMsg, End, Sendme, Truncated};
+use tor_cell::relaycell::extend::{CcRequest, CircRequestExt};
+use tor_cell::relaycell::msg::{AnyRelayMsg, End, Sendme, SendmeTag, Truncated};
 use tor_cell::relaycell::{
     AnyRelayMsgOuter, RelayCellDecoder, RelayCellDecoderResult, RelayCellFormat, RelayCmd,
     StreamId, UnparsedRelayMsg,
@@ -55,8 +55,10 @@ use oneshot_fused_workaround as oneshot;
 use safelog::sensitive as sv;
 use tracing::{debug, trace, warn};
 
+#[cfg(feature = "conflux")]
+use super::conflux::ConfluxMsgHandler;
 use super::{
-    CellHandlers, CircuitHandshake, CloseStreamBehavior, ReactorResultChannel, SendRelayCell,
+    CellHandlers, CircuitHandshake, CloseStreamBehavior, LegId, ReactorResultChannel, SendRelayCell,
 };
 
 use std::borrow::Borrow;
@@ -64,6 +66,7 @@ use std::pin::Pin;
 use std::result::Result as StdResult;
 use std::sync::{Arc, Mutex};
 use std::task::Poll;
+use std::time::{Duration, SystemTime};
 
 use create::{Create2Wrap, CreateFastWrap, CreateHandshakeWrap};
 use extender::HandshakeAuxDataHandler;
@@ -74,6 +77,12 @@ use {
     tor_cell::relaycell::msg::Begin,
 };
 
+#[cfg(feature = "conflux")]
+use {
+    super::conflux::{ConfluxAction, OooRelayMsg},
+    crate::tunnel::reactor::RemoveLegReason,
+};
+
 /// Initial value for outbound flow-control window on streams.
 pub(super) const SEND_WINDOW_INIT: u16 = 500;
 /// Initial value for inbound flow-control window on streams.
@@ -147,15 +156,19 @@ pub(crate) struct Circuit {
     crypto_out: OutboundClientCrypt,
     /// List of hops state objects used by the reactor
     hops: Vec<CircHop>,
-    /// Mutable information about this circuit, shared with
-    /// [`ClientCirc`](crate::circuit::ClientCirc).
-    ///
-    // TODO(conflux)/TODO(#1840): this belongs in the Reactor
-    mutable: Arc<Mutex<MutableState>>,
+    /// Mutable information about this circuit,
+    /// shared with the reactor's `ConfluxSet`.
+    mutable: Arc<MutableState>,
     /// This circuit's identifier on the upstream channel.
     channel_id: CircId,
     /// An identifier for logging about this reactor's circuit.
     unique_id: UniqId,
+    /// A handler for conflux cells.
+    ///
+    /// Set once the conflux handshake is initiated by the reactor
+    /// using [`Reactor::handle_link_circuits`](super::Reactor::handle_link_circuits).
+    #[cfg(feature = "conflux")]
+    conflux_handler: Option<ConfluxMsgHandler>,
     /// Memory quota account
     #[allow(dead_code)] // Partly here to keep it alive as long as the circuit
     memquota: CircuitAccount,
@@ -190,10 +203,55 @@ pub(super) enum CircuitCmd {
         /// The reason for closing the stream.
         reason: streammap::TerminateReason,
     },
+    /// Remove this circuit from the conflux set.
+    ///
+    /// Returned by `ConfluxMsgHandler::handle_conflux_msg` for invalid messages
+    /// (originating from wrong hop), and for messages that are rejected
+    /// by its inner `AbstractMsgHandler`.
+    #[cfg(feature = "conflux")]
+    ConfluxRemove(RemoveLegReason),
+    /// This circuit has completed the conflux handshake,
+    /// and wants to send the specified cell.
+    ///
+    /// Returned by an `AbstractMsgHandler` to signal to the reactor that
+    /// the conflux handshake is complete.
+    #[cfg(feature = "conflux")]
+    ConfluxHandshakeComplete(SendRelayCell),
     /// Perform a clean shutdown on this circuit.
     CleanShutdown,
+    /// Enqueue an out-of-order cell in the reactor.
+    #[cfg(feature = "conflux")]
+    Enqueue(OooRelayMsg),
 }
 
+/// Return a `CircProto` error for the specified unsupported cell.
+///
+/// This error will shut down the reactor.
+///
+/// Note: this is a macro to simplify usage (this way the caller doesn't
+/// need to .map() the result to the appropriate type)
+macro_rules! unsupported_client_cell {
+    ($msg:expr) => {{
+        unsupported_client_cell!(@ $msg, "")
+    }};
+
+    ($msg:expr, $hopnum:expr) => {{
+        let hop: HopNum = $hopnum;
+        let hop_display = format!(" from hop {}", hop.display());
+        unsupported_client_cell!(@ $msg, hop_display)
+    }};
+
+    (@ $msg:expr, $hopnum_display:expr) => {
+        Err(crate::Error::CircProto(format!(
+            "Unexpected {} cell{} on client circuit",
+            $msg.cmd(),
+            $hopnum_display,
+        )))
+    };
+}
+
+pub(super) use unsupported_client_cell;
+
 impl Circuit {
     /// Create a new non-multipath circuit.
     pub(super) fn new(
@@ -202,7 +260,7 @@ impl Circuit {
         unique_id: UniqId,
         input: CircuitRxReceiver,
         memquota: CircuitAccount,
-        mutable: Arc<Mutex<MutableState>>,
+        mutable: Arc<MutableState>,
     ) -> Self {
         let chan_sender = SometimesUnboundedSink::new(channel.sender());
 
@@ -217,10 +275,77 @@ impl Circuit {
             channel_id,
             crypto_out,
             mutable,
+            #[cfg(feature = "conflux")]
+            conflux_handler: None,
             memquota,
         }
     }
 
+    /// Return the process-unique identifier of this circuit.
+    pub(super) fn unique_id(&self) -> UniqId {
+        self.unique_id
+    }
+
+    /// Return the shared mutable state of this circuit.
+    pub(super) fn mutable(&self) -> &Arc<MutableState> {
+        &self.mutable
+    }
+
+    /// Install a [`ConfluxMsgHandler`] on this circuit,
+    ///
+    /// Once this is called, the circuit will be able to handle conflux cells.
+    #[cfg(feature = "conflux")]
+    pub(super) fn install_conflux_handler(&mut self, conflux_handler: ConfluxMsgHandler) {
+        self.conflux_handler = Some(conflux_handler);
+    }
+
+    /// Send a LINK cell to the specified hop.
+    ///
+    /// This must be called *after* a [`ConfluxMsgHandler`] is installed
+    /// on the circuit with [`install_conflux_handler`](Self::install_conflux_handler).
+    #[cfg(feature = "conflux")]
+    pub(super) async fn begin_conflux_link(
+        &mut self,
+        hop: HopNum,
+        cell: AnyRelayMsgOuter,
+        runtime: &tor_rtcompat::DynTimeProvider,
+    ) -> Result<()> {
+        use tor_rtcompat::SleepProvider as _;
+
+        if self.conflux_handler.is_none() {
+            return Err(internal!(
+                "tried to send LINK cell before installing a ConfluxMsgHandler?!"
+            )
+            .into());
+        }
+
+        let cell = SendRelayCell {
+            hop,
+            early: false,
+            cell,
+        };
+        self.send_relay_cell(cell).await?;
+
+        let Some(conflux_handler) = self.conflux_handler.as_mut() else {
+            return Err(internal!("ConfluxMsgHandler disappeared?!").into());
+        };
+
+        Ok(conflux_handler.note_link_sent(runtime.wallclock())?)
+    }
+
+    /// Get the wallclock time when the handshake on this circuit is supposed to time out.
+    ///
+    /// Returns `None` if this handler hasn't started the handshake yet.
+    pub(super) fn conflux_hs_timeout(&self) -> Option<SystemTime> {
+        cfg_if::cfg_if! {
+            if #[cfg(feature = "conflux")] {
+                self.conflux_handler.as_ref().map(|handler| handler.handshake_timeout())?
+            } else {
+                None
+            }
+        }
+    }
+
     /// Handle a [`CtrlMsg::AddFakeHop`](super::CtrlMsg::AddFakeHop) message.
     #[cfg(test)]
     pub(super) fn handle_add_fake_hop(
@@ -263,7 +388,7 @@ impl Circuit {
         hop: HopNum,
         early: bool,
         msg: AnyRelayMsgOuter,
-    ) -> Result<(AnyChanMsg, &[u8; SENDME_TAG_LEN])> {
+    ) -> Result<(AnyChanMsg, SendmeTag)> {
         let mut body: RelayCellBody = msg
             .encode(relay_format, &mut rand::rng())
             .map_err(|e| Error::from_cell_enc(e, "relay cell body"))?
@@ -291,6 +416,11 @@ impl Circuit {
     ///
     /// Does not check whether the cell is well-formed or reasonable.
     pub(super) async fn send_relay_cell(&mut self, msg: SendRelayCell) -> Result<()> {
+        if self.is_conflux_pending() {
+            // TODO(conflux): is this right? Should we ensure all the legs are linked?
+            return Err(internal!("tried to send cell on unlinked circuit").into());
+        }
+
         let SendRelayCell {
             hop,
             early,
@@ -321,6 +451,12 @@ impl Circuit {
                 ent.take_capacity_to_send(msg.msg())?;
             }
         }
+
+        // Save the RelayCmd of the message before it gets consumed below.
+        // We need this to tell our ConfluxMsgHandler about the cell we've just sent,
+        // so that it can update its counters.
+        let relay_cmd = msg.cmd();
+
         // NOTE(eta): Now that we've encrypted the cell, we *must* either send it or abort
         //            the whole circuit (e.g. by returning an error).
         let (msg, tag) =
@@ -328,12 +464,17 @@ impl Circuit {
         // The cell counted for congestion control, inform our algorithm of such and pass down the
         // tag for authenticated SENDMEs.
         if c_t_w {
-            circhop.ccontrol.note_data_sent(tag)?;
+            circhop.ccontrol.note_data_sent(&tag)?;
         }
 
         let cell = AnyChanCell::new(Some(self.channel_id), msg);
         Pin::new(&mut self.chan_sender).send_unbounded(cell).await?;
 
+        #[cfg(feature = "conflux")]
+        if let Some(conflux) = self.conflux_handler.as_mut() {
+            conflux.note_cell_sent(relay_cmd);
+        }
+
         Ok(())
     }
 
@@ -352,12 +493,13 @@ impl Circuit {
     pub(super) fn handle_cell(
         &mut self,
         handlers: &mut CellHandlers,
+        leg: LegId,
         cell: ClientCircChanMsg,
     ) -> Result<Vec<CircuitCmd>> {
         trace!("{}: handling cell: {:?}", self.unique_id, cell);
         use ClientCircChanMsg::*;
         match cell {
-            Relay(r) => self.handle_relay_cell(handlers, r),
+            Relay(r) => self.handle_relay_cell(handlers, leg, r),
             Destroy(d) => {
                 let reason = d.reason();
                 debug!(
@@ -377,7 +519,7 @@ impl Circuit {
     fn decode_relay_cell(
         &mut self,
         cell: Relay,
-    ) -> Result<(HopNum, CircTag, RelayCellDecoderResult)> {
+    ) -> Result<(HopNum, SendmeTag, RelayCellDecoderResult)> {
         // This is always RELAY, not RELAY_EARLY, so long as this code is client-only.
         let cmd = cell.cmd();
         let mut body = cell.into_relay_body().into();
@@ -385,15 +527,6 @@ impl Circuit {
         // Decrypt the cell. If it's recognized, then find the
         // corresponding hop.
         let (hopnum, tag) = self.crypto_in.decrypt(cmd, &mut body)?;
-        // Make a copy of the authentication tag. TODO: I'd rather not
-        // copy it, but I don't see a way around it right now.
-        let tag = {
-            let mut tag_copy = [0_u8; SENDME_TAG_LEN];
-            // TODO(nickm): This could crash if the tag length changes.  We'll
-            // have to refactor it then.
-            tag_copy.copy_from_slice(tag);
-            tag_copy
-        };
 
         // Decode the cell.
         let decode_res = self
@@ -408,13 +541,14 @@ impl Circuit {
             .decode(body.into())
             .map_err(|e| Error::from_bytes_err(e, "relay cell"))?;
 
-        Ok((hopnum, tag.into(), decode_res))
+        Ok((hopnum, tag, decode_res))
     }
 
     /// React to a Relay or RelayEarly cell.
     fn handle_relay_cell(
         &mut self,
         handlers: &mut CellHandlers,
+        leg: LegId,
         cell: Relay,
     ) -> Result<Vec<CircuitCmd>> {
         let (hopnum, tag, decode_res) = self.decode_relay_cell(cell)?;
@@ -439,7 +573,7 @@ impl Circuit {
             // that SendmeEmitMinVersion is no more than 1.  If the authorities
             // every increase that parameter to a higher number, this will
             // become incorrect.  (Higher numbers are not currently defined.)
-            let sendme = Sendme::new_tag(tag.into());
+            let sendme = Sendme::from(tag);
             let cell = AnyRelayMsgOuter::new(None, sendme.into());
             circ_cmds.push(CircuitCmd::Send(SendRelayCell {
                 hop: hopnum,
@@ -461,7 +595,7 @@ impl Circuit {
 
         let (mut msgs, incomplete) = decode_res.into_parts();
         while let Some(msg) = msgs.next() {
-            let msg_status = self.handle_relay_msg(handlers, hopnum, c_t_w, msg)?;
+            let msg_status = self.handle_relay_msg(handlers, hopnum, leg, c_t_w, msg)?;
 
             match msg_status {
                 None => continue,
@@ -496,6 +630,7 @@ impl Circuit {
         &mut self,
         handlers: &mut CellHandlers,
         hopnum: HopNum,
+        leg: LegId,
         cell_counts_toward_windows: bool,
         msg: UnparsedRelayMsg,
     ) -> Result<Option<CircuitCmd>> {
@@ -509,6 +644,54 @@ impl Circuit {
             return self.handle_meta_cell(handlers, hopnum, msg);
         };
 
+        #[cfg(feature = "conflux")]
+        let msg = if let Some(conflux) = self.conflux_handler.as_mut() {
+            match conflux.action_for_msg(hopnum, cell_counts_toward_windows, streamid, msg)? {
+                ConfluxAction::Deliver(msg) => {
+                    // The message either doesn't count towards the sequence numbers
+                    // or is already well-ordered, so we're ready to handle it.
+
+                    // It's possible that some of our buffered messages are now ready to be
+                    // handled. We don't check that here, however, because that's handled
+                    // by the reactor main loop.
+                    msg
+                }
+                ConfluxAction::Enqueue(msg) => {
+                    // Tell the reactor to enqueue this msg
+                    return Ok(Some(CircuitCmd::Enqueue(msg)));
+                }
+            }
+        } else {
+            // If we don't have a conflux_handler, it means this circuit is not part of
+            // a conflux tunnel, so we can just process the message.
+            msg
+        };
+
+        self.handle_in_order_relay_msg(
+            handlers,
+            hopnum,
+            leg,
+            cell_counts_toward_windows,
+            streamid,
+            msg,
+        )
+    }
+
+    /// Handle a single incoming relay message that is known to be in order.
+    pub(super) fn handle_in_order_relay_msg(
+        &mut self,
+        handlers: &mut CellHandlers,
+        hopnum: HopNum,
+        leg: LegId,
+        cell_counts_toward_windows: bool,
+        streamid: StreamId,
+        msg: UnparsedRelayMsg,
+    ) -> Result<Option<CircuitCmd>> {
+        #[cfg(feature = "conflux")]
+        if let Some(conflux) = self.conflux_handler.as_mut() {
+            conflux.inc_last_seq_delivered(&msg);
+        }
+
         let hop = self
             .hop_mut(hopnum)
             .ok_or_else(|| Error::CircProto("Cell from nonexistent hop!".into()))?;
@@ -535,8 +718,7 @@ impl Circuit {
                 // response
                 hop_map.ending_msg_received(streamid)?;
                 drop(hop_map);
-                // TODO(conflux): we will need a new CircuitCmd::SendOnPrimaryLeg for this
-                return self.handle_incoming_stream_request(handlers, msg, streamid, hopnum);
+                return self.handle_incoming_stream_request(handlers, msg, streamid, hopnum, leg);
             }
             Some(StreamEntMut::EndSent(EndSentStreamEnt { half_stream, .. })) => {
                 // We sent an end but maybe the other side hasn't heard.
@@ -555,7 +737,7 @@ impl Circuit {
             ) =>
             {
                 drop(hop_map);
-                return self.handle_incoming_stream_request(handlers, msg, streamid, hopnum);
+                return self.handle_incoming_stream_request(handlers, msg, streamid, hopnum, leg);
             }
             _ => {
                 // No stream wants this message, or ever did.
@@ -567,6 +749,60 @@ impl Circuit {
         Ok(None)
     }
 
+    /// Handle a conflux message coming from the specified hop.
+    ///
+    /// Returns an error if
+    ///
+    ///   * this is not a conflux circuit (i.e. it doesn't have a [`ConfluxMsgHandler`])
+    ///   * this is a client circuit and the conflux message originated an unexpected hop
+    ///   * the cell was sent in violation of the handshake protocol
+    #[cfg(feature = "conflux")]
+    fn handle_conflux_msg(
+        &mut self,
+        hop: HopNum,
+        msg: UnparsedRelayMsg,
+    ) -> Result<Option<CircuitCmd>> {
+        let Some(conflux_handler) = self.conflux_handler.as_mut() else {
+            // If conflux is not enabled, tear down the circuit
+            // (see 4.2.1. Cell Injection Side Channel Mitigations in prop329)
+            //
+            // TODO(conflux): make sure this is properly implemented
+            return Err(Error::CircProto(format!(
+                "Received {} cell from hop {} on non-conflux client circuit?!",
+                msg.cmd(),
+                hop.display(),
+            )));
+        };
+
+        Ok(conflux_handler.handle_conflux_msg(msg, hop))
+    }
+
+    /// For conflux: return the sequence number of the last cell sent on this leg.
+    ///
+    /// Returns an error if this circuit is not part of a conflux set.
+    #[cfg(feature = "conflux")]
+    pub(super) fn last_seq_sent(&self) -> Result<u64> {
+        let handler = self
+            .conflux_handler
+            .as_ref()
+            .ok_or_else(|| internal!("tried to get last_seq_sent of non-conflux circ"))?;
+
+        Ok(handler.last_seq_sent())
+    }
+
+    /// For conflux: return the sequence number of the last cell received on this leg.
+    ///
+    /// Returns an error if this circuit is not part of a conflux set.
+    #[cfg(feature = "conflux")]
+    pub(super) fn last_seq_recv(&self) -> Result<u64> {
+        let handler = self
+            .conflux_handler
+            .as_ref()
+            .ok_or_else(|| internal!("tried to get last_seq_recv of non-conflux circ"))?;
+
+        Ok(handler.last_seq_recv())
+    }
+
     /// Deliver `msg` to the specified open stream entry `ent`.
     fn deliver_msg_to_stream(
         streamid: StreamId,
@@ -620,6 +856,7 @@ impl Circuit {
         msg: UnparsedRelayMsg,
         stream_id: StreamId,
         hop_num: HopNum,
+        leg: LegId,
     ) -> Result<Option<CircuitCmd>> {
         use super::syncview::ClientCircSyncView;
         use tor_cell::relaycell::msg::EndReason;
@@ -733,6 +970,7 @@ impl Circuit {
             req,
             stream_id,
             hop_num,
+            leg,
             msg_tx,
             receiver,
             memquota,
@@ -875,7 +1113,7 @@ impl Circuit {
 
         let relay_cell_format = cell_protocol.relay_cell_format();
         let BoxedClientLayer { fwd, back, binding } =
-            cell_protocol.construct_layers(HandshakeRole::Initiator, keygen)?;
+            cell_protocol.construct_client_layers(HandshakeRole::Initiator, keygen)?;
 
         trace!("{}: Handshake complete; circuit created.", self.unique_id);
 
@@ -967,36 +1205,7 @@ impl Circuit {
         let relay_cell_protocol = RelayCryptLayerProtocol::Tor1(RelayCellFormat::V0);
 
         // Set the client extensions.
-        // allow 'unused_mut' because of the combinations of `cfg` conditions below
-        #[allow(unused_mut)]
-        let mut client_extensions = Vec::new();
-
-        if params.ccontrol.is_enabled() {
-            cfg_if::cfg_if! {
-                if #[cfg(feature = "flowctl-cc")] {
-                    // TODO(arti#88): We have an `if false` in `exit_circparams_from_netparams`
-                    // which should prevent the above `is_enabled()` from ever being true,
-                    // even with the "flowctl-cc" feature enabled:
-                    // https://gitlab.torproject.org/tpo/core/arti/-/merge_requests/2932#note_3191196
-                    // The panic here is so that CI tests will hopefully catch if congestion
-                    // control is unexpectedly enabled.
-                    // We should remove this panic once xon/xoff flow is supported.
-                    #[cfg(not(test))]
-                    panic!("Congestion control is enabled on this circuit, but we don't yet support congestion control");
-
-                    #[allow(unreachable_code)]
-                    client_extensions.push(NtorV3Extension::RequestCongestionControl);
-                } else {
-                    return Err(
-                        internal!(
-                            "Congestion control is enabled on this circuit, but 'flowctl-cc' feature is not enabled"
-                        )
-                        .into()
-                    );
-                }
-            }
-        }
-
+        let client_extensions = circ_extensions_from_params(params)?;
         let wrap = Create2Wrap {
             handshake_type: HandshakeType::NTOR_V3,
         };
@@ -1042,14 +1251,13 @@ impl Circuit {
         self.hops.push(hop);
         self.crypto_in.add_layer(rev);
         self.crypto_out.add_layer(fwd);
-        let mut mutable = self.mutable.lock().expect("poisoned lock");
-        Arc::make_mut(&mut mutable.path).push_hop(peer_id);
-        mutable.binding.push(binding);
+        self.mutable.add_hop(peer_id, binding);
 
         Ok(())
     }
 
     /// Handle a RELAY cell on this circuit with stream ID 0.
+    #[allow(clippy::cognitive_complexity)]
     fn handle_meta_cell(
         &mut self,
         handlers: &mut CellHandlers,
@@ -1095,8 +1303,34 @@ impl Circuit {
 
         trace!("{}: Received meta-cell {:?}", self.unique_id, msg);
 
+        #[cfg(feature = "conflux")]
+        if matches!(
+            msg.cmd(),
+            RelayCmd::CONFLUX_LINK
+                | RelayCmd::CONFLUX_LINKED
+                | RelayCmd::CONFLUX_LINKED_ACK
+                | RelayCmd::CONFLUX_SWITCH
+        ) {
+            return self.handle_conflux_msg(hopnum, msg);
+        }
+
+        if self.is_conflux_pending() {
+            warn!(
+                "{}: received unexpected cell {msg:?} on unlinked conflux circuit",
+                self.unique_id,
+            );
+            return Err(Error::CircProto(
+                "Received unexpected cell on unlinked circuit".into(),
+            ));
+        }
+
         // For all other command types, we'll only get them in response
         // to another command, which should have registered a responder.
+        //
+        // TODO:(conflux): should the conflux state machine be a meta cell handler?
+        // We'd need to add support for multiple meta handlers, and change the
+        // MetaCellHandler API to support returning Option<RunOnceCmdInner>
+        // (because some cells will require sending a response)
         if let Some(mut handler) = handlers.meta_handler.take() {
             if handler.expected_hop() == hopnum {
                 // Somebody was waiting for a message -- maybe this message
@@ -1121,19 +1355,13 @@ impl Circuit {
                 // Somebody wanted a message from a different hop!  Put this
                 // one back.
                 handlers.meta_handler = Some(handler);
-                Err(Error::CircProto(format!(
-                    "Unexpected {} cell from hop {} on client circuit",
-                    msg.cmd(),
-                    hopnum.display(),
-                )))
+
+                unsupported_client_cell!(msg, hopnum)
             }
         } else {
             // No need to call shutdown here, since this error will
             // propagate to the reactor shut it down.
-            Err(Error::CircProto(format!(
-                "Unexpected {} cell on client circuit",
-                msg.cmd()
-            )))
+            unsupported_client_cell!(msg)
         }
     }
 
@@ -1150,15 +1378,10 @@ impl Circuit {
             .hop_mut(hopnum)
             .ok_or_else(|| Error::CircProto(format!("Couldn't find hop {}", hopnum.display())))?;
 
-        let tag = match msg.into_tag() {
-            Some(v) => CircTag::try_from(v.as_slice())
-                .map_err(|_| Error::CircProto("malformed tag on circuit sendme".into()))?,
-            None => {
+        let tag = msg.into_sendme_tag().ok_or_else(||
                 // Versions of Tor <=0.3.5 would omit a SENDME tag in this case;
                 // but we don't support those any longer.
-                return Err(Error::CircProto("missing tag on circuit sendme".into()));
-            }
-        };
+                 Error::CircProto("missing tag on circuit sendme".into()))?;
         // Update the CC object that we received a SENDME along with possible congestion signals.
         hop.ccontrol.note_sendme_received(tag, signals)?;
         Ok(None)
@@ -1322,6 +1545,17 @@ impl Circuit {
         Ok(())
     }
 
+    /// Returns true if there are any streams on this circuit
+    ///
+    /// Important: this function locks the stream map of its each of the [`CircHop`]s
+    /// in this circuit, so it must **not** be called from any function where the
+    /// stream map lock is held (such as [`ready_streams_iterator`](Self::ready_streams_iterator).
+    pub(super) fn has_streams(&self) -> bool {
+        self.hops
+            .iter()
+            .any(|hop| hop.map.lock().expect("lock poisoned").n_open_streams() > 0)
+    }
+
     /// The number of hops in this circuit.
     pub(super) fn num_hops(&self) -> u8 {
         // `Circuit::add_hop` checks to make sure that we never have more than `u8::MAX` hops,
@@ -1339,12 +1573,23 @@ impl Circuit {
         !self.hops.is_empty()
     }
 
+    /// Get the `HopNum` of the last hop, if this circuit is non-empty.
+    ///
+    /// Returns `None` if the circuit has no hops.
+    pub(super) fn last_hop_num(&self) -> Option<HopNum> {
+        let num_hops = self.num_hops();
+        if num_hops == 0 {
+            // asked for the last hop, but there are no hops
+            return None;
+        }
+        Some(HopNum::from(num_hops - 1))
+    }
+
     /// Get the path of the circuit.
     ///
     /// **Warning:** Do not call while already holding the [`Self::mutable`] lock.
     pub(super) fn path(&self) -> Arc<path::Path> {
-        let mutable = self.mutable.lock().expect("poisoned lock");
-        Arc::clone(&mutable.path)
+        self.mutable.path()
     }
 
     /// Return a ClockSkew declaring how much clock skew the other side of this channel
@@ -1360,6 +1605,49 @@ impl Circuit {
         let hop = self.hop(hop)?;
         Some(hop.ccontrol.uses_stream_sendme())
     }
+
+    /// Returns whether this is a conflux circuit that is not linked yet.
+    pub(super) fn is_conflux_pending(&self) -> bool {
+        let Some(status) = self.conflux_status() else {
+            return false;
+        };
+
+        status != ConfluxStatus::Linked
+    }
+
+    /// Returns the conflux status of this circuit.
+    ///
+    /// Returns `None` if this is not a conflux circuit.
+    pub(super) fn conflux_status(&self) -> Option<ConfluxStatus> {
+        cfg_if::cfg_if! {
+            if #[cfg(feature = "conflux")] {
+                self.conflux_handler
+                    .as_ref()
+                    .map(|handler| handler.status())
+            } else {
+                None
+            }
+        }
+    }
+
+    /// Returns initial RTT on this leg, measured in the conflux handshake.
+    #[cfg(feature = "conflux")]
+    pub(super) fn init_rtt(&self) -> Option<Duration> {
+        self.conflux_handler
+            .as_ref()
+            .map(|handler| handler.init_rtt())?
+    }
+}
+
+/// The conflux status of a conflux [`Circuit`].
+#[derive(Copy, Clone, Debug, PartialEq, Eq)]
+pub(super) enum ConfluxStatus {
+    /// Circuit has not begun the conflux handshake yet.
+    Unlinked,
+    /// Conflux handshake is in progress.
+    Pending,
+    /// A linked conflux circuit.
+    Linked,
 }
 
 /// Return the stream ID of `msg`, if it has one.
@@ -1485,7 +1773,7 @@ impl CircHop {
 
     /// Delegate to CongestionControl, for testing purposes
     #[cfg(test)]
-    pub(crate) fn send_window_and_expected_tags(&self) -> (u32, Vec<CircTag>) {
+    pub(crate) fn send_window_and_expected_tags(&self) -> (u32, Vec<SendmeTag>) {
         self.ccontrol.send_window_and_expected_tags()
     }
 
@@ -1496,4 +1784,46 @@ impl CircHop {
     pub(super) fn n_open_streams(&self) -> usize {
         self.map.lock().expect("lock poisoned").n_open_streams()
     }
+
+    /// Return a reference to our CongestionControl object.
+    pub(crate) fn ccontrol(&self) -> &CongestionControl {
+        &self.ccontrol
+    }
+}
+
+/// Return the client circuit-creation extensions that we should use in order to negotiate
+/// a given set of circuit hop parameters.
+#[allow(clippy::unnecessary_wraps)]
+pub(super) fn circ_extensions_from_params(params: &CircParameters) -> Result<Vec<CircRequestExt>> {
+    // allow 'unused_mut' because of the combinations of `cfg` conditions below
+    #[allow(unused_mut)]
+    let mut client_extensions = Vec::new();
+
+    if params.ccontrol.is_enabled() {
+        cfg_if::cfg_if! {
+            if #[cfg(feature = "flowctl-cc")] {
+                // TODO(arti#88): We have an `if false` in `exit_circparams_from_netparams`
+                // which should prevent the above `is_enabled()` from ever being true,
+                // even with the "flowctl-cc" feature enabled:
+                // https://gitlab.torproject.org/tpo/core/arti/-/merge_requests/2932#note_3191196
+                // The panic here is so that CI tests will hopefully catch if congestion
+                // control is unexpectedly enabled.
+                // We should remove this panic once xon/xoff flow is supported.
+                #[cfg(not(test))]
+                panic!("Congestion control is enabled on this circuit, but we don't yet support congestion control");
+
+                #[allow(unreachable_code)]
+                client_extensions.push(CircRequestExt::CcRequest(CcRequest::default()));
+            } else {
+                return Err(
+                    tor_error::internal!(
+                        "Congestion control is enabled on this circuit, but 'flowctl-cc' feature is not enabled"
+                    )
+                    .into()
+                );
+            }
+        }
+    }
+
+    Ok(client_extensions)
 }
