@@ -11,17 +11,14 @@ use crate::crypto::handshake::ntor_v3::{NtorV3Client, NtorV3PublicKey};
 use crate::stream::AnyCmdChecker;
 use crate::tunnel::circuit::celltypes::CreateResponse;
 use crate::tunnel::circuit::{path, CircParameters};
+use crate::tunnel::reactor::circuit::circ_extensions_from_params;
 use crate::tunnel::reactor::{NtorClient, ReactorError};
-use crate::tunnel::{streammap, HopLocation, LegId, TargetHop};
+use crate::tunnel::{streammap, HopLocation, TargetHop};
 use crate::util::skew::ClockSkew;
 use crate::Result;
 use tor_cell::chancell::msg::HandshakeType;
-use tor_cell::relaycell::extend::NtorV3Extension;
 use tor_cell::relaycell::msg::{AnyRelayMsg, Sendme};
-use tor_cell::relaycell::{
-    AnyRelayMsgOuter, RelayCellFormat, RelayCellFormatTrait, RelayCellFormatV0, StreamId,
-    UnparsedRelayMsg,
-};
+use tor_cell::relaycell::{AnyRelayMsgOuter, RelayCellFormat, StreamId, UnparsedRelayMsg};
 use tor_error::{bad_api_usage, into_bad_api_usage, Bug};
 use tracing::trace;
 #[cfg(feature = "hs-service")]
@@ -31,10 +28,10 @@ use {
 };
 
 #[cfg(test)]
-use crate::congestion::sendme::CircTag;
+use tor_cell::relaycell::msg::SendmeTag;
 
 #[cfg(feature = "conflux")]
-use super::Circuit;
+use super::{Circuit, ConfluxLinkResultChannel};
 
 use oneshot_fused_workaround as oneshot;
 
@@ -43,7 +40,6 @@ use crate::tunnel::circuit::{StreamMpscReceiver, StreamMpscSender};
 use tor_linkspec::{EncodedLinkSpec, OwnedChanTarget};
 
 use std::result::Result as StdResult;
-use std::sync::Arc;
 
 /// A message telling the reactor to do something.
 ///
@@ -176,6 +172,23 @@ pub(crate) enum CtrlMsg {
         /// Oneshot channel to return the clock skew.
         answer: oneshot::Sender<StdResult<ClockSkew, Bug>>,
     },
+    /// Link the specified circuits into the current tunnel,
+    /// to form a multi-path tunnel.
+    #[cfg(feature = "conflux")]
+    #[allow(unused)] // TODO(conflux)
+    LinkCircuits {
+        /// The circuits to link into the tunnel,
+        #[educe(Debug(ignore))]
+        circuits: Vec<Circuit>,
+        /// Oneshot channel to notify sender when all the specified circuits have finished linking,
+        /// or have failed to link.
+        ///
+        /// A client circuit is said to be fully linked once the `RELAY_CONFLUX_LINKED_ACK` is sent
+        /// (see [set construction]).
+        ///
+        /// [set construction]: https://spec.torproject.org/proposals/329-traffic-splitting.html#set-construction
+        answer: ConfluxLinkResultChannel,
+    },
 }
 
 /// A message telling the reactor to do something.
@@ -226,11 +239,6 @@ pub(crate) enum CtrlCmd {
         #[cfg(feature = "hs-service")]
         filter: Box<dyn IncomingStreamRequestFilter>,
     },
-    /// Get the leg ID and path for each leg of the tunnel.
-    QueryLegs {
-        /// Oneshot channel to notify on completion.
-        done: ReactorResultChannel<Vec<(LegId, Arc<path::Path>)>>,
-    },
     /// (tests only) Add a hop to the list of hops on this circuit, with dummy cryptography.
     #[cfg(test)]
     AddFakeHop {
@@ -244,7 +252,7 @@ pub(crate) enum CtrlCmd {
     #[cfg(test)]
     QuerySendWindow {
         hop: HopNum,
-        done: ReactorResultChannel<(u32, Vec<CircTag>)>,
+        done: ReactorResultChannel<(u32, Vec<SendmeTag>)>,
     },
     /// Shut down the reactor, and return the underlying [`Circuit`],
     /// if the tunnel is not multi-path.
@@ -306,7 +314,7 @@ impl<'a> ControlHandler<'a> {
                 params,
                 done,
             } => {
-                let Ok((_id, circ)) = self.reactor.circuits.single_leg_mut() else {
+                let Ok((leg_id, circ)) = self.reactor.circuits.single_leg_mut() else {
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot extend multipath tunnel"
@@ -316,27 +324,26 @@ impl<'a> ControlHandler<'a> {
                     return Ok(None);
                 };
 
-                // ntor handshake only supports V0.
-                /// Local type alias to ensure consistency below.
-                type Rcf = RelayCellFormatV0;
-
-                let (extender, cell) =
-                    CircuitExtender::<NtorClient, Tor1RelayCrypto<Rcf>, _, _>::begin(
-                        Rcf::FORMAT,
-                        peer_id,
-                        HandshakeType::NTOR,
-                        &public_key,
-                        linkspecs,
-                        params,
-                        &(),
-                        circ,
-                        done,
-                    )?;
+                let (extender, cell) = CircuitExtender::<NtorClient, Tor1RelayCrypto, _, _>::begin(
+                    RelayCellFormat::V0,
+                    peer_id,
+                    HandshakeType::NTOR,
+                    &public_key,
+                    linkspecs,
+                    params,
+                    &(),
+                    circ,
+                    done,
+                )?;
                 self.reactor
                     .cell_handlers
                     .set_meta_handler(Box::new(extender))?;
 
-                Ok(Some(RunOnceCmdInner::Send { cell, done: None }))
+                Ok(Some(RunOnceCmdInner::Send {
+                    leg: leg_id,
+                    cell,
+                    done: None,
+                }))
             }
             CtrlMsg::ExtendNtorV3 {
                 peer_id,
@@ -345,7 +352,7 @@ impl<'a> ControlHandler<'a> {
                 params,
                 done,
             } => {
-                let Ok((_id, circ)) = self.reactor.circuits.single_leg_mut() else {
+                let Ok((leg_id, circ)) = self.reactor.circuits.single_leg_mut() else {
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot extend multipath tunnel"
@@ -356,43 +363,13 @@ impl<'a> ControlHandler<'a> {
                 };
 
                 // TODO #1067, TODO #1947: support negotiating other formats.
-                /// Local type alias to ensure consistency below.
-                type Rcf = RelayCellFormatV0;
+                let relay_cell_format = RelayCellFormat::V0;
 
-                // Set the client extensions.
-                // allow 'unused_mut' because of the combinations of `cfg` conditions below
-                #[allow(unused_mut)]
-                let mut client_extensions = Vec::new();
-
-                if params.ccontrol.is_enabled() {
-                    cfg_if::cfg_if! {
-                        if #[cfg(feature = "flowctl-cc")] {
-                            // TODO(arti#88): We have an `if false` in `exit_circparams_from_netparams`
-                            // which should prevent the above `is_enabled()` from ever being true,
-                            // even with the "flowctl-cc" feature enabled:
-                            // https://gitlab.torproject.org/tpo/core/arti/-/merge_requests/2932#note_3191196
-                            // The panic here is so that CI tests will hopefully catch if congestion
-                            // control is unexpectedly enabled.
-                            // We should remove this panic once xon/xoff flow is supported.
-                            #[cfg(not(test))]
-                            panic!("Congestion control is enabled on this circuit, but we don't yet support congestion control");
-
-                            #[allow(unreachable_code)]
-                            client_extensions.push(NtorV3Extension::RequestCongestionControl);
-                        } else {
-                            return Err(
-                                tor_error::internal!(
-                                    "Congestion control is enabled on this circuit, but 'flowctl-cc' feature is not enabled"
-                                )
-                                .into()
-                            );
-                        }
-                    }
-                }
+                let client_extensions = circ_extensions_from_params(&params)?;
 
                 let (extender, cell) =
-                    CircuitExtender::<NtorV3Client, Tor1RelayCrypto<Rcf>, _, _>::begin(
-                        Rcf::FORMAT,
+                    CircuitExtender::<NtorV3Client, Tor1RelayCrypto, _, _>::begin(
+                        relay_cell_format,
                         peer_id,
                         HandshakeType::NTOR_V3,
                         &public_key,
@@ -406,7 +383,11 @@ impl<'a> ControlHandler<'a> {
                     .cell_handlers
                     .set_meta_handler(Box::new(extender))?;
 
-                Ok(Some(RunOnceCmdInner::Send { cell, done: None }))
+                Ok(Some(RunOnceCmdInner::Send {
+                    leg: leg_id,
+                    cell,
+                    done: None,
+                }))
             }
             CtrlMsg::BeginStream {
                 hop,
@@ -448,6 +429,7 @@ impl<'a> ControlHandler<'a> {
 
                 let cell = circ.begin_stream(hop_num, message, sender, rx, cmd_checker)?;
                 Ok(Some(RunOnceCmdInner::BeginStream {
+                    leg: leg_id,
                     cell,
                     hop: hop_location,
                     done,
@@ -511,17 +493,18 @@ impl<'a> ControlHandler<'a> {
                     early: false,
                     cell,
                 };
-                // TODO(conflux): We need to pass the leg id in `RunOnceCmdInner::Send`, but we
-                // can't add this to `RunOnceCmdInner` until the stream map knows about legs.
-                // `Reactor::ready_streams_iterator` now knows the leg id, so we should be able to
-                // use that.
+
                 Ok(Some(RunOnceCmdInner::Send {
+                    leg: leg_id,
                     cell,
                     done: Some(sender),
                 }))
             }
             // TODO(conflux): this should specify which leg to send the msg on
-            // (currently we send it down the primary leg)
+            // (currently we send it down the primary leg).
+            //
+            // This will involve updating ClientCIrc::send_raw_msg() to take a
+            // leg id argument (which is a breaking change.
             #[cfg(feature = "send-control-msg")]
             CtrlMsg::SendMsg {
                 hop_num,
@@ -534,7 +517,11 @@ impl<'a> ControlHandler<'a> {
                     early: false,
                     cell,
                 };
+
+                let leg = self.reactor.circuits.primary_leg_id();
+
                 Ok(Some(RunOnceCmdInner::Send {
+                    leg,
                     cell,
                     done: Some(sender),
                 }))
@@ -554,6 +541,10 @@ impl<'a> ControlHandler<'a> {
             CtrlMsg::FirstHopClockSkew { answer } => {
                 Ok(Some(RunOnceCmdInner::FirstHopClockSkew { answer }))
             }
+            #[cfg(feature = "conflux")]
+            CtrlMsg::LinkCircuits { circuits, answer } => {
+                Ok(Some(RunOnceCmdInner::Link { circuits, answer }))
+            }
         }
     }
 
@@ -616,17 +607,6 @@ impl<'a> ControlHandler<'a> {
 
                 Ok(())
             }
-            CtrlCmd::QueryLegs { done } => {
-                let ret = self
-                    .reactor
-                    .circuits
-                    .legs()
-                    .map(|(id, leg)| (id, leg.path()))
-                    .collect();
-                // TODO(conflux): Consider adding the leg id to the `Path`.
-                let _ = done.send(Ok(ret));
-                Ok(())
-            }
             #[cfg(test)]
             CtrlCmd::AddFakeHop {
                 relay_cell_format,
